@@ -94,6 +94,7 @@
 #'   ## Example 1: AP/TM reconstruction on a symmetric 20x20 matrix
 #'   ## (10 percent known entries)
 #'
+#'   RNGkind("L'Ecuyer-CMRG")
 #'   set.seed(123456789)
 #'
 #'   m <- 20L
@@ -160,6 +161,7 @@
 #'   ## with zero diagonal and reduced (20,20) submodels
 #'   ## (20 percent known entries)
 #'
+#'   RNGkind("L'Ecuyer-CMRG")
 #'   set.seed(123456789)
 #'
 #'   m <- 40L
@@ -263,27 +265,27 @@ tmpinv <- function(S=NULL, M=NULL, b_row=NULL, b_col=NULL, b_val=NULL,
   bounds   <- lapply(bounds, function(v)   {
     if (length(v) != 2L)     stop("Each bounds element must have length 2.")
     vapply(v, function(x)    if (is.null(x) || is.na(x) || length(x) != 1L )
-    NA_real_                 else           as.numeric(x), numeric(1L))
+      NA_real_                 else           as.numeric(x), numeric(1L))
   })
   if (all(vapply(bounds,     function(v)    all(is.na(v)), logical(1L))))
     return(result)                                     # finish if unbounded
-
+  
   # (result) Perform bound-constrained iterative refinement
   b_lim      <- matrix(rbind(vapply(bounds, function(v)  if (is.na(v[2]))  Inf
                                     else                           v[2],
                                     numeric(1L)),
                              vapply(bounds, function(v)  if (is.na(v[1])) -Inf
                                     else                           v[1],
-                                    numeric(1L))),          ncol=      1)
+                                    numeric(1L))),          ncol=     1L)
   C_lim      <- matrix(rbind(diag(n_cells), diag(n_cells)), ncol=n_cells)
   S          <- if (!is.null(S)) S  else    matrix(0,  nrow=length(b_row) +
-                                                       length(b_col),
-                                                       ncol=0)
+                                                     length(b_col),
+                                                   ncol=0L)
   S          <- rbind(cbind(S, matrix(0, nrow=nrow(S), ncol=      2 * n_cells)),
                       cbind(   matrix(0, nrow=n_cells, ncol=          ncol(S)),
                                diag(n_cells),
                                matrix(0, nrow=n_cells, ncol=          n_cells)),
-                     -cbind(   matrix(0, nrow=n_cells, ncol=ncol(S) + n_cells),
+                      -cbind(  matrix(0, nrow=n_cells, ncol=ncol(S) + n_cells),
                                diag(n_cells)))
   finite_rows   <- is.finite(b_lim[, 1L])              # drop rows with +-np.inf
   nonzero_cols  <- colSums(abs(S[c(rep(TRUE, nrow(S) - length(finite_rows)),
@@ -292,7 +294,7 @@ tmpinv <- function(S=NULL, M=NULL, b_row=NULL, b_col=NULL, b_val=NULL,
   for   (iter in seq_len(iteration_limit)) {
     M_idx  <- integer(0)
     b_val  <- numeric(0)
-    x      <- matrix(result$x, ncol=1)
+    x      <- matrix(result$x, ncol=1L)
     for (k in seq_len(n_cells)) {
       if ((!is.na(bounds[[k]][1]) && x[k] < bounds[[k]][1] - tolerance) ||
           (!is.na(bounds[[k]][2]) && x[k] > bounds[[k]][2] + tolerance)) {
@@ -304,36 +306,37 @@ tmpinv <- function(S=NULL, M=NULL, b_row=NULL, b_col=NULL, b_val=NULL,
     if (length(M_idx) < n_cells) {
       M      <- diag(n_cells)[M_idx, , drop=FALSE]
       result <- suppressWarnings(do.call(.tmpinv.instance,
-                        c(list(b_row=b_row, b_col=b_col,  b_val=b_val,
-                               b_lim=b_lim[finite_rows, , drop=FALSE], i=i, j=j,
-                               C_lim=C_lim[finite_rows, , drop=FALSE],
-                               S=S[c(rep(TRUE, nrow(S) -  nrow(b_lim)),
-                                     finite_rows),        nonzero_cols,
-                                   drop=FALSE],           M=M,
-                               zero_diagonal=zero_diagonal, reduced=reduced,
-                               symmetric=symmetric, tolerance=tolerance,
-                               iteration_limit=iteration_limit, r=r,
-                               final=final, alpha=alpha),
-                          dots)))
+                                         c(list(b_row=b_row, b_col=b_col,  b_val=b_val,
+                                                b_lim=b_lim[finite_rows, , drop=FALSE], i=i, j=j,
+                                                C_lim=C_lim[finite_rows, , drop=FALSE],
+                                                S=S[c(rep(TRUE, nrow(S) -  nrow(b_lim)),
+                                                      finite_rows),        nonzero_cols,
+                                                    drop=FALSE],           M=M,
+                                                zero_diagonal=zero_diagonal, reduced=reduced,
+                                                symmetric=symmetric, tolerance=tolerance,
+                                                iteration_limit=iteration_limit, r=r,
+                                                final=final, alpha=alpha),
+                                           dots)))
     } else {
       break
     }
   }
-
+  
   # (result) Replace out-of-bound values with replace_value
-  x        <- matrix(result$x, ncol=1)
+  x        <- matrix(result$x, ncol=1L)
   x_lb     <- vapply(bounds, function(v) if (is.na(v[1])) -Inf
                      else                          v[1], numeric(1L))
   x_ub     <- vapply(bounds, function(v) if (is.na(v[2]))  Inf
                      else                          v[2], numeric(1L))
   x[(x < x_lb - tolerance) | (x > x_ub + tolerance)] <- replace_value
-  result$x <- matrix(x, nrow=nrow(result$x),  ncol=ncol(result$x))
-
+  result$x <- matrix(x, nrow=nrow(result$x),  ncol=ncol(result$x),
+                     byrow=TRUE)
+  
   # enforce (final) symmetry
   if (isTRUE(symmetric)) if (nrow(result$x) == ncol(result$x))
     result$x <-          0.5 *   (result$x  +     t(result$x))         else
       stop("symmetric=True requires a square matrix (m == p).")
-
+  
   result
 }
 #' @export
@@ -379,8 +382,8 @@ print.summary.tmpinv <- function(x, ...) {
     stop("Minimum length for b_row and b_col is 2.")
   if (!all(is.finite(b_row)) || !all(is.finite(b_col)))
     stop("b_row and b_col must not contain +-Inf or NA")
-  b_row <- matrix(as.numeric(b_row), ncol=1L)
-  b_col <- matrix(as.numeric(b_col), ncol=1L)
+  b_row <- matrix(as.numeric(b_row), ncol=1L, byrow=TRUE)
+  b_col <- matrix(as.numeric(b_col), ncol=1L, byrow=TRUE)
   m     <- nrow(b_row) * i
   p     <- nrow(b_col) * j
   if (!is.null(S))                    {
@@ -400,7 +403,7 @@ print.summary.tmpinv <- function(x, ...) {
   if (!is.null(b_val))                {
     if (is.null(M))               stop("Both M and b_val must be defined." )
     if (!all(is.finite(b_val)))   stop("b_val must not contain +-Inf or NA.")
-    b_val <- matrix(b_val, ncol=1L)
+    b_val <- matrix(b_val, ncol=1L, byrow=TRUE)
   }
   if (!is.null(M) && !is.null(b_val)) {
     if (nrow(M) != nrow(b_val))   stop(sprintf(paste0("M and b_val must have ",
@@ -408,7 +411,7 @@ print.summary.tmpinv <- function(x, ...) {
                                                       "rows: %d vs %d"),
                                                nrow(M), nrow(b_val)))
   }
-
+  
   # perform full estimation and return the result
   if (is.null(reduced)) {
     result       <- list(full=TRUE,                   model=NULL,
@@ -418,15 +421,15 @@ print.summary.tmpinv <- function(x, ...) {
                                                                  ncol=1L))
     if (!is.null(b_val))      b_blocks <- rbind(b_blocks, matrix(b_val,
                                                                  ncol=1L))
-    b            <- matrix(b_blocks,      ncol=1)
+    b            <- matrix(b_blocks,      ncol=1L)
     result$model <- do.call(rclsp::clsp, c(list(problem="ap", b=b, C=dots$C_lim,
                                                 S=S, M=M, m=m, p=p, i=i, j=j,
                                                 zero_diagonal=zero_diagonal),
                                            dots[setdiff(names(dots),
                                                         c("C_lim", "b_lim"))]))
     result$x     <- result$model$x
-
-  # perform reduced estimation and return the result
+    
+    # perform reduced estimation and return the result
   } else {
     reduced      <- as.integer(reduced)
     if (length(reduced)  != 2L                                               ||
@@ -446,9 +449,11 @@ print.summary.tmpinv <- function(x, ...) {
       b_val  <-  if (is.null(b_val) || length(b_val) == 0L) b_diag     else
         rbind(b_val, b_diag)
       M      <-  if (is.null(M)     || length(M)     == 0L) M_diag     else
-        rbind(M,     M_diag)
-      b_val  <-  b_val[!duplicated(M), ,        drop=FALSE]
-      M      <-      M[!duplicated(M), ,        drop=FALSE]
+        rbind(M,      M_diag)
+      ord    <- do.call(order, as.data.frame(M))
+      idx    <- ord[!duplicated(M[ord, , drop=FALSE])]
+      M      <- M[idx, ,                 drop=FALSE]
+      b_val  <- b_val[idx,               drop=FALSE]
     }
     if (!is.null(M))           {
       if (!(all(        abs(M    ) <   sqrt(.Machine$double.eps)  |
@@ -458,7 +463,7 @@ print.summary.tmpinv <- function(x, ...) {
         stop(paste0("M must be a unique row subset ",
                     "of the identity matrix in the ",
                     "reduced model."))
-      X_true   <- matrix(NA_real_,     nrow=m, ncol=p)
+      X_true   <- matrix(NA_real_,     nrow=1L*m, ncol=p)
       for (idx in seq_len(nrow(M))) {
         col          <- which.max(M[idx, ])
         r            <- (col - 1L) %/% p  + 1L         # M has m * p columns
@@ -466,7 +471,30 @@ print.summary.tmpinv <- function(x, ...) {
         X_true[r, c] <- b_val[idx]
       }
     }
-    if (!is.null(S))           {
+    if (!is.null(dots$b_lim))  {
+      X_lim    <- matrix(NA_real_,     nrow=2L*m, ncol=p)
+      l_idx    <- if (nrow(pos<-which(
+        S[(nrow(S)-nrow(dots$b_lim)+1L):nrow(S), ] == -1,
+        arr.ind=TRUE)) > 0) pos[1,1] else nrow(dots$b_lim)
+      if (l_idx > 1L               ) {
+        for (idx in 1:(l_idx - 1L        )) {
+          col  <- which.max(dots$C_lim[idx, ])
+          r    <- floor((col - 1L) /  p) + 1L
+          c    <-      ((col - 1L) %% p) + 1L
+          X_lim[  r, c] <-  dots$b_lim[idx  ]
+        }
+      }
+      if (l_idx <= nrow(dots$b_lim)) {
+        for (idx in l_idx:nrow(dots$C_lim)) {
+          col  <- which.max(dots$C_lim[idx, ])
+          r    <- floor((col - 1L) /  p) + 1L
+          c    <-      ((col - 1L) %% p) + 1L
+          X_lim[m+r, c] <-  dots$b_lim[idx  ]
+        }
+      }
+    }
+    if (!is.null(S) && any(S[1:(if (is.null(dots$b_lim)) nrow(S) else nrow(S)-
+                                nrow(dots$b_lim)), ] != 0)) {
       warning("User-provided S is ignored in the reduced model.", call.=FALSE)
     }
     for   (row_block in seq_len(ceiling(m / m_subset))) {
@@ -475,20 +503,44 @@ print.summary.tmpinv <- function(x, ...) {
         m_end    <- min(m_start + m_subset - 1L,   m)
         p_start  <- (col_block  - 1L) * p_subset + 1L
         p_end    <- min(p_start + p_subset - 1L,   p)
+        C_subset <-  NULL
         S_subset <- diag((m_end - m_start  + 1L) +    (p_end - p_start  + 1L))
         b_subset <- matrix(c(b_row[m_start:m_end],
                              b_col[p_start:p_end]),                  ncol=1L)
         M_subset <-  NULL
         if (!is.null(M))      {
           subset     <- matrix(X_true[m_start:m_end, p_start:p_end], ncol=1L)
-          non_empty  <- !is.na(subset)
+          non_empty  <- !is.na(subset) & is.finite(subset)
           if (any(non_empty)) {
             M_subset <- diag(length(subset))[non_empty, , drop=FALSE]
             b_subset <- matrix(c(b_subset, subset[non_empty]),       ncol=1L)
           }
         }
+        if (!is.null(dots$b_lim))  {
+          subset     <- as.vector(t(rbind(
+            X_lim[(m_start  ):(m_end  ), p_start:p_end],
+            X_lim[(m+m_start):(m+m_end), p_start:p_end]
+          )))
+          non_empty  <- !is.na(subset) & is.finite(subset)
+          if (any(non_empty)) {
+            b_subset <- matrix(c(b_subset, subset[non_empty]),       ncol=1L)
+            C_subset <- rbind(diag(h <- length(subset) %/% 2),
+                              diag(h)      )[non_empty, , drop=FALSE]
+            S        <- cbind(rbind(diag(h),
+                                    matrix(0, nrow=h, ncol=h))[non_empty, ,
+                                                               drop=FALSE],
+                              rbind(matrix(0, nrow=h, ncol=h),
+                                    -diag(h)                  )[non_empty, ,
+                                                                drop=FALSE])
+            S_subset <- rbind(cbind(S_subset, matrix(0, nrow=nrow(S_subset),
+                                                     ncol=ncol(S) )),
+                              cbind(matrix(0, nrow=nrow(S),
+                                           ncol=ncol(S_subset)), S))
+          }
+        }
         tmp      <-  do.call(rclsp::clsp, c(list(problem="ap", b=b_subset,
-                                                 S=S_subset,   M=M_subset,
+                                                 C=C_subset,   S=S_subset,
+                                                 M=M_subset,
                                                  m=m_end - m_start + 1L,
                                                  p=p_end - p_start + 1L,
                                                  i=i, j=j, zero_diagonal=FALSE),
@@ -499,7 +551,7 @@ print.summary.tmpinv <- function(x, ...) {
       }
     }
   }
-
+  
   # enforce symmetry
   if (isTRUE(symmetric)) if (nrow(result$x) == ncol(result$x))
     result$x    <-       0.5 *   (result$x  +     t(result$x))         else
